/* JRDemo.prg
   JRound() is a function which may solve rounding problems as well
   as inaccuracies in certain math functions.

   The syntax for JRound() is:

          JRound( <nNumber> [, <nDecimals>] )

   JRound() increments the least significant bit of the Mantissa
   in the equation nNumber := ( 2 ^ Exponent ) * Mantissa. GO MSKB and search
   IEEE and Floating Point for explanations of double-precision numbers.

   If <nDecimals> is NIL, the lsb is incremented once.

   If <nDecimals> is not NIL, and not 98, the lsb of the Mantissa is
   incremented, a new number id created by ROUND() and the lsb of the
   Mantissa of new number is then incremented.

   If <nDecimals> is 98, the lsb is incremented twice. This is required
   to get the correct result from some math functions such as
   log(n)/log(base); e.g., log(10^9)/log(10) returns 8.999999999999998
   instead of 9.0. Note that accuracy may be obtained in division operations
   by using JRound( Numerator ); i.e., JRound( log(10^9) ) / log(10).

   Basically, the 16th significant decimal position is affected with
   this function. Please give it a workout with your own range of numbers
   and send feedback to Jo W. French 74730,1751.

   Compile with /n. Link Blinker fi jrdemo lib JFRound
*/


FUNCTION Main()
  #define PAUSE  QOUT() ; WAIT ; CLS ; QOUT(c)
  LOCAL aR, i, nTest := -117.57 * 573.50, c
  aR := {1.265, 146.795,  8.245,  18.435,  10.165,  16.685, ;
         1.025,   4.645,  8.155,  35.105, 286.335,   4.515, 4.225, ;
         117.57 * 573.50, 162.23 / 2 }
  CLS

  ? "Display of JRound results showing the result as a number, as a"
  ? "string of 64 bits and the ASC() values of an 8 byte float. Note"
  ? "that the displayed byte order is the reverse of an F2Bin result."
  ? "-----  This number is the product of -117.57 * 573.50.  -------"
  ? XSTR( nTest )
  ? c := JBinStr( nTest )
  ? YDisp( c )
  ? "-----  This is the result of JRound( number ).  ---------------"
  ? XSTR( JRound(nTest) )
  ? c := JBinStr( JRound(nTest) )
  ? YDisp( c )
  ? "-----  This is the result of JRound( number, 98 ).  -----------"
  ? XSTR( JRound(nTest,98) )
  ? c := JBinStr( JRound(nTest,98) )
  ? YDisp( c )
  ? "-----  This is the result of Round( JRound( number ), 2 ).  ---"
  ? XSTR( ROUND(JRound(nTest),2) )
  ? c := JBinStr( ROUND(JRound(nTest),2) )
  ? YDisp( c )
  ? "-----  This is the result of JRound( number, 2 ).  ------------"
  ? XSTR( JRound(nTest,2) )
  ? c := JBinStr( JRound(nTest,2) )
  ? YDisp( c )

  c := "       Number                   Round( Number, 2 )" + ;
       "       JRound( Number, 2 )"
  PAUSE

  AEVAL( aR, {|a| QOUT( XSTR(a) + ;
                        XSTR(ROUND(a,2)) + ;
                        XSTR(JRound( a, 2 ) ) ) } )
  PAUSE

  AEVAL( aR, {|a| QOUT( XSTR(-a) + ;
                        XSTR(ROUND(-a,2)) + ;
                        XSTR(JRound(-a, 2 ) ) ) } )

  c := "     Number = LOG10( 10^n)       INT( Number ) " + ;
       "       INT(JRound( Number, 98 ))"
  PAUSE

  FOR i := 0 TO 31
      nTest := LOG10( 10 ^ i )
      ?  XSTR( nTest ) + XSTR( INT( nTest ) ) + ;
                         XSTR( INT( JRound( nTest,98 ) ) )
      If i == 15
         PAUSE
      Endif
  NEXT

  PAUSE

  FOR i := 0 TO 31
      nTest := LOG10( 10 ^ -i )
      ? XSTR( nTest ) + XSTR( INT( nTest ) ) + ;
                        XSTR( INT( JRound( nTest, 98 ) ) )
      If i == 15
         PAUSE
      Endif
  NEXT

  c := "   Number .05735 * .011797      Round( Number, 10 )" + ;
       "       JRound( Number, 10 )"

  PAUSE

  nTest := .05735 * .011797
  ? XSTR( nTest ) + XSTR(  ROUND( nTest, 10) ) + ;
                    XSTR( JRound( nTest, 10 ) )
  ?
  ? "   Number 555735.05 * 555735.70  Round( Number, 2 ) " + ;
    "       JRound( Number, 2 )"
  nTest := 555735.05 * 555735.70
  ? STR( nTest,26, 5  ) + STR(  ROUND( nTest, 2 ), 26, 5 ) + ;
                          STR( JRound( nTest, 2 ), 26, 5 )
  ?
RETURN NIL

FUNCTION XSTR( n )
RETURN STR( n, 26, 19 )

FUNCTION JBinStr( nNumber )
   LOCAL cBinary   := IF( nNumber < 0, "1", "0" ), i        // Sign Bit
   LOCAL nExponent := Exponent( nNumber ) + 1022
   LOCAL nMantissa := ABS( Mantissa( nNumber )) * 2 - 1
   IF nNumber != 0

      FOR i := 10 TO 0 step -1
          cBinary += IF( INT(nExponent / 2^i ) > 0, "1","0" )
          nExponent %= 2^i
      NEXT

      FOR i := -1 to -52 step -1
          cBinary += IF( INT( nMantissa / 2^i ) > 0, "1","0" )
          nMantissa %= 2^i
      NEXT
   ELSE
      cBinary := REPLICATE( "0", 64 )
   ENDIF

RETURN cBinary

FUNCTION YDisp( c )
  LOCAL cTmp, i, j, n, cRet := ""
  FOR i := 1 TO 57 STEP 8
      cTmp := SUBSTR( c, i, 8 )
      n := 0
      FOR j := 0 TO 7
          IF SUBSTR( cTmp, j + 1, 1 ) == "1"
             n += 2 ^ ( 7 - j )
          ENDIF
      NEXT
      cRet += STR(n,8)
  NEXT
RETURN cRet

FUNCTION LOG10( nNumber )
RETURN LOG( nNumber ) / LOG( 10 )

